#pragma warning(disable:4786)
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <string>
#include "Spreadsheet.h"
#include "Function.h"
#include "Parser.h"

// Test driver and main loop.

using namespace std;

void testParser()
{
    string str("A1*(B2+B3)");  
    Tokenizer t(str.begin(),str.end());
    string token;
    Tokenizer::TokenType type;
    while (t.getNext(token,type))
    {
        cout << "Token: " << token << ", type = " << type << endl;
    }

    Spreadsheet s;
    s.at("A1").setValue(1);
    s.at("B2").setValue(2);
    s.at("B3").setValue(3);
    Parser p;
    Value v = p.evalString(str, &s);
}

void testCell()
{
    // A basic test which works even if
    // Spreadsheet and Function aren't
    // implemented yet:
    Cell c1(NULL);
    c1.setValue(42.89);
    Cell c2(NULL);
    c1.subscribe(&c2);
    c1.setValue(19); // Should call update()
    c1.printOn(cout);
}

void testSpreadsheet()
{
    // Test the spreadsheet without
    // functions or dependencies:

    Spreadsheet s;
    s.setSize(3,3);
    s.at_rowcol(1,1).setValue(42.6);
    s.at_rowcol(1,2).setValue(13.9);
    s.printOn(cout);
}

void testAll()
{
    // Test the spreadsheet with
    // functions enabled:

    const Value v1 = 42.6;
    const Value v2 = 58.23;
    const Value v3 = 23.1;

    Spreadsheet s;
    s.setSize(3,3);
    s.at_rowcol(1,1).setValue(v1);
    s.at_rowcol(1,2).setValue(v2);

    // When we set a value, it should
    // stay set:
    assert(s.at_rowcol(1,1).getValue() == v1);
    assert(s.at_rowcol(1,2).getValue() == v2);
}

void initSpreadsheet(Spreadsheet &s)
{
    // Set up the formula:
    // ((a+b) / (a*b)) - ((a+b) % (a*b))
    //
    // Cell assignments:
    //  1,1 : a
    //  1,2 : b
    //  2,1 : a+b
    //  2,2 : a*b
    //  3,1 : (a+b)/(a*b)
    //  3,2 : (a+b)%(a*b)
    //  4,1 : ((a+b) / (a*b)) - ((a+b) % (a*b))
}

bool doCmd(Spreadsheet &s)
{
    // Execute one command and display the spreadsheet.
    cout << "->";

    string cmd;
    cin >> cmd;
    if (cmd == "set")
    {
        string addr;
        cin >> addr;
        string newContents;
        cin >> newContents;

        // Create a new cell and set its
        // contents:
        Cell c(&s);
        c.setFunc(newContents);
        s.at(addr) = c;
    }
    else if (cmd == "show") // Re-show spreadsheet
    {
        // No-op, we always show down below.    
    }
    else if (cmd == "showfuncs") // Show with functions
    {
        cout << "====================" << endl;
        s.printOn(cout,true);
        cout << endl;
        return true;
    }
    else if (cmd == "quit")
        return false; // Exit the program
    else
    {
        cerr << "Unknown command: " << cmd << endl;
        cerr << "Available commands:\nset <cell> <formula/value>\nshow\nshowfuncs]quit" << endl;
        return true;
    }
    cout << "--------------------" << endl;
    s.printOn(cout);
    cout << endl;
    return true; // Keep accepting commands
}


void run()
{
    // Initialize the spreadsheet to a hard-coded
    // function and enter an input loop to set
    // the values:
    Spreadsheet s;
    initSpreadsheet(s);

    bool keepGoing;
    do 
    {
        try 
        {
            keepGoing = doCmd(s);
        }
        catch (SyntaxError &e)
        {
            cout << "Syntax error: " << e.what() << endl;
        }
        catch (SpreadsheetError &e)
        {
            cout << "Error: " << e.what() << endl;
        }
        catch (...)
        {
            cout << "Fatal error: Unknown exception" << endl;
            throw;
        }
    } while (keepGoing);
}


main()
{
    // Following tests can be re-enabled
    // for regression testing at any time:

    //testParser();
    //testCell();
    //testSpreadsheet();
    //testAll();

    run();
}
