#pragma warning(disable:4786)
#include <iostream>
#include <algorithm>
#include <cmath>
#include "Parser.h"
#include "Cell.h"
#include "ICellContainer.h"
using namespace std;

Tokenizer::Tokenizer(TextIter begin, TextIter end)
: mPos(begin), mEnd(end)
{}


bool Tokenizer::getNext(string &token, TokenType &type, bool remove)
{
    // Hard coded syntax:
    // Whitespace separates tokens and is
    //   otherwise ignored.
    // Operators terminate tokens and are
    //   tokens themselves.

    // Table of known operators:
    static const char OPS[] = "+-*/%,()";
    static const char *OPS_END = OPS+sizeof(OPS);

    TextIter p = mPos;
    // Skip any leading whitespace:
    while (p != mEnd && isspace(*p))
        ++p;
        
    TextIter startTok = p;

    // Check to see if we're on an operator; if
    // so, that's our token:    
    if (p != mEnd && (find(OPS,OPS_END,*p)!=OPS_END))
    {
        switch (*p)
        {
        case '(': type = LPAREN; break;
        case ')': type = RPAREN; break;
        default: type = OPERATOR; break;
        }
        ++p;
    }
    else
    {
        type = ERROR;
        while (p != mEnd)
        {
            if (type==ERROR && isdigit(*p)) type=CONSTANT;
            if (type==ERROR && isalpha(*p)) type=NAME;
            if (type==NAME && isdigit(*p)) type=CELLREF;
            if (isspace(*p)) break;
            if (find(OPS,OPS_END,*p)!=OPS_END)
                break;
            // Anything other than a space or
            // a known operator char
            // is taken to be part of the token.
            ++p;
        }
    }

    // Return the token found; this will be
    // empty if mPos = end (meaning no more
    // tokens exist).  An empty token is a
    // signal to stop parsing.
    token = string(startTok,p);

    if (remove)
    {
        // At this point, p is pointing to one-past-
        // the-end of the current token.
        mPos = p; // Remember for next time
    }

    return !token.empty();
}

//------------------------------------------------
Value Parser::evalLeaf(string const &token, Tokenizer::TokenType type)
{
    switch (type)
    {
    case Tokenizer::CONSTANT: 
        return atof(token.c_str());
        break;
    case Tokenizer::CELLREF: 
        {
            Cell const *cell = mCellContainer->findCell(token);
            if (cell)
                return cell->getValue();
            else
                return 0.0; // Default for an undefined cell
        }
        break;
    default:
        throw SyntaxError(token);
    }
}

Value Parser::evalOp(char op)
{
    Value lhs = popArg();
    Value rhs = popArg();
    cout << "Calculating " << lhs << op << rhs << endl; 
    switch (op)
    {
    case '+': return lhs+rhs; break;
    case '-': return lhs-rhs; break;
    case '*': return lhs*rhs; break;
    case '/': return lhs/rhs; break;
    case '%':
        {
            return lhs - floor(lhs/rhs)*rhs;
        }
        break;
    default:
        throw SyntaxError(string("Unknown operator ")+string(op,1)); 
    }
}

Value Parser::evalString(string const &s, ICellContainer const *container)
{
    mCellContainer = container;
    Tokenizer tokenizer(s.begin(),s.end());
    return eval(tokenizer);
}

Value Parser::eval(Tokenizer &tok)
{
    string token;
    Tokenizer::TokenType type;

    while (tok.getNext(token,type))
    {
        switch (type)
        {
        case Tokenizer::ERROR: throw SyntaxError(token); break;
        case Tokenizer::CONSTANT: case Tokenizer::CELLREF: 
            cout << "Pushing leaf: " << token << endl;
            argStack.push_back(evalLeaf(token,type)); break;
            break;
        case Tokenizer::OPERATOR:
            {
                char op = token[0];
                switch (op)
                {
                case '+': case '-': 
                    // Evaluate the rest of the string and push it back
                    argStack.push_back(eval(tok));
                    argStack.push_back(evalOp(op));
                    break;
                case '*': case '/': case '%':
                    // Use greedy parsing:
                    if (!tok.getNext(token,type))
                        throw SyntaxError("No right hand argument to operator *");
                    if (type==Tokenizer::LPAREN)
                        argStack.push_back(eval(tok));
                    else if (type==Tokenizer::NAME)
                        handleFuncCall(tok, token);
                    else
                        argStack.push_back(evalLeaf(token,type));
                    argStack.push_back(evalOp(op));
                    break;
                case ',':
                    // Used to push arguments to functions on the stack.
                    // Left hand side has been pushed; swallow the ,
                    // operator so the next arg will be pushed.
                    break;
                default:
                    throw SyntaxError(token);
                    break;
                }
            }
            break;

        case Tokenizer::RPAREN:
            // End of subexpression
            return popArg();
            break;

        case Tokenizer::LPAREN:
            // Start of subexpression
            cout << "Pushing subexpression result" << endl;
            argStack.push_back(eval(tok));
            break;
        case Tokenizer::NAME:
            // Start of a function call
            handleFuncCall(tok, token);
            break;
        default:
            throw SyntaxError("Cannot parse string");
        }
    }

    return popArg();
}

void Parser::handleFuncCall(Tokenizer &tok, std::string funcName)
{
    string token;
    Tokenizer::TokenType type;
    // Look up the function in the collection of installed
    // functions:
    map<string,FuncAdapter*>::iterator it = installedFuncs.find(funcName);
    if (it == installedFuncs.end())
        throw SyntaxError(string("Unknown function ")+token);
    FuncAdapter *fa = it->second;
    
    int numArgs = fa->getNumArgs();
    // Next token should be a '(':
    if (!tok.getNext(token,type) || type != Tokenizer::LPAREN) 
        throw SyntaxError(token);
    if (numArgs==0)
    {
        if (!tok.getNext(token,type) || type != Tokenizer::RPAREN)
            throw SyntaxError(token);
    }
    else
    {
        // Check number of arguments passed:
        int startStackLevel = argStack.size();
        Value lastArg = eval(tok);
        argStack.push_back(lastArg); // Push back what RPAREN popped
        if (argStack.size() - startStackLevel != fa->getNumArgs())
            throw SyntaxError(string("Wrong # of args to ")+fa->getName());
    }                
    // We have the right number of arguments, so
    // call the function:
    argStack.push_back(fa->eval(argStack));
}


Value Parser::popArg()
{
    if (argStack.size()==0)
        throw SyntaxError("One or more arguments missing");
    Value v = argStack.back();
    argStack.pop_back();
    return v;
}

//static 
FuncAdapter *Parser::install(FuncAdapter *fa)
{
    installedFuncs[fa->getName()] = fa;
    return fa;
}

std::map<std::string,FuncAdapter*> Parser::installedFuncs;

//--------------------------------------------------------------------
// Functions in standard C library:
static FuncAdapter *f_sin = Parser::install(makeAdapter("sin",sin));
static FuncAdapter *f_cos = Parser::install(makeAdapter("cos",cos));
static FuncAdapter *f_tan = Parser::install(makeAdapter("tan",tan));
static FuncAdapter *f_pow = Parser::install(makeAdapter("pow",pow)); // 2 args
static FuncAdapter *f_floor = Parser::install(makeAdapter("floor",floor));

// Other clients can call Parser::install to install their own
// functions.