#ifndef PARSER_H
#define PARSER_H

#include <string>
#include <list>
#include <exception>
#include <map>
class ICellContainer;
#include "Defs.h"

typedef std::string::const_iterator TextIter;

class SyntaxError : public SpreadsheetError
{
public:
    SyntaxError(std::string const &m) : SpreadsheetError(m) {}
};

// A class that takes care of tokenizing an input
// stream.
class Tokenizer
{
public:
    Tokenizer(TextIter begin,
              TextIter end);

    typedef enum {ERROR=0, CONSTANT, CELLREF, OPERATOR, LPAREN, RPAREN, NAME} TokenType;
    bool getNext(std::string &token, TokenType &type, bool remove=true);

private:
    TextIter mPos;
    TextIter mEnd;
};

class FuncAdapter;

// A class that takes care of parsing
// spreadsheet formulas.
class Parser
{
public:
    Value evalString(std::string const &s, ICellContainer const *container);

    static FuncAdapter *install(FuncAdapter *fa);
private:
    ICellContainer const *mCellContainer;
    std::list<Value> argStack;

    Value eval(Tokenizer &tok);
    Value evalOp(char op);
    Value evalLeaf(std::string const &token, Tokenizer::TokenType type);
    Value popArg();
    void handleFuncCall(Tokenizer &tok, std::string funcName);

    static std::map<std::string,FuncAdapter*> installedFuncs;
};

// An interface class that adapts global C++ functions to
// the Parser interface:
class FuncAdapter
{
public:
    virtual std::string getName() const = 0;
    virtual int getNumArgs() const =0;
    virtual Value eval(std::list<Value> &argStack) const = 0;
};


// A class template that helps automate the process of
// installing a new function adapter.  It adapts functions
// of up to two arguments of types A1 and A2 to which Value
// can convert, and returns a type Res which can be converted
// to Value.
template <typename Res, typename A1=Value, typename A2=Value>
class TFuncAdapter : public FuncAdapter
{
public:
    typedef Res (*FUNC0)();
    typedef Res (*FUNC1)(A1);
    typedef Res (*FUNC2)(A1,A2);
    TFuncAdapter(std::string n, FUNC0 f) : f0(f), numArgs(0), name(n) {}
    TFuncAdapter(std::string n, FUNC1 f) : f1(f), numArgs(1), name(n) {}
    TFuncAdapter(std::string n, FUNC2 f) : f2(f), numArgs(2), name(n) {}

    // String name of the function:
    virtual std::string getName() const {return name;}

    // Number of arguments of the function:
    virtual int getNumArgs() const {return numArgs;}

    // Evaluate the function to a single value, given an
    // argument stack:
    virtual Value eval(std::list<Value> &argStack) const 
    {
        switch (numArgs)
        {
        case 0: return f0(); break;
        case 1: return f1(popArg(argStack)); break;
        case 2: return f2(popArg(argStack), popArg(argStack)); break;
        }
        return 0;
    }
   
private:
    Value popArg(std::list<Value> &argStack) const
    {
        if (argStack.size()==0) 
            throw SyntaxError(string("Not enough arguments for ")+name);
        Value v = argStack.back();
        argStack.pop_back();
        return v;
    }

    // The underlying function pointer, appropriately typed:
    union 
    {
        FUNC0 f0;
        FUNC1 f1;
        FUNC2 f2;
    };
    const int numArgs;      // Number of arguments the function accepts
    const std::string name; // Name of the function
};

// The following helper functions make it a bit easier to install
// new functions; they take care of automatically deducing the
// correct template arguments, given an appropriate function
// pointer.
template <typename Res>
inline
TFuncAdapter<Res> *makeAdapter(std::string const &n, Res (*f)()) {
    return new TFuncAdapter<Res>(n,f);
}

template <typename Res, typename A1>
inline
TFuncAdapter<Res,A1> *makeAdapter(std::string const &n, Res (*f)(A1)) {
    return new TFuncAdapter<Res,A1>(n,f);
}

template <typename Res, typename A1, typename A2>
inline
TFuncAdapter<Res,A1,A2> *makeAdapter(std::string const &n, Res (*f)(A1,A2)) {
    return new TFuncAdapter<Res,A1,A2>(n,f);
}

// Sample usage:  
// Parser::install(makeAdapter("sin",sin));

#endif 