#pragma warning(disable:4786)
#include <iostream>
#include <iomanip>
#include <cassert>
#include <string>
#include "Spreadsheet.h"
#include "Function.h"
using namespace std;

const Value Cell::NULL_VALUE = 0.0; //1.0/0.0; //numeric_limits<Value>::infinity() ;

Cell::Cell(ICellContainer *cnt) 
: cachedValue(NULL_VALUE), container(cnt), inUpdate(false)
{}

Cell::~Cell()
{
    clear();
}

Cell::Cell(Cell const &other) 
    : container(other.container), inUpdate(false)
{
    // Subscribe this cell to the appropriate sources:
    setFunc(other.formula.getText());
}

Cell &Cell::operator=(Cell const &other)
{
    // Subscribe this cell to the appropriate sources:
    container = other.container;
    inUpdate = false;
    setFunc(other.formula.getText());
    return *this;
}

void Cell::setContainer(ICellContainer *cnt)
{
    container=cnt;
    setFunc(formula.getText()); // Force re-subscribe
}

void Cell::update(const Publisher *)
{
    auto_resetter<bool> ars(inUpdate); // Auto-reset to false on exit

    if (inUpdate)
       throw SpreadsheetError("Loop detected, stopping calculation");
    inUpdate = true;
        
    cachedValue = NULL_VALUE;
    cachedValue = formula.evaluate(container);
    notify();
}

void Cell::setValue(const Value &v)
{
    // Clear the cell of any associated
    // function, set the cached value,
    // and notify dependents about
    // the change:
    char buf[32];
    ltoa(v,buf,10);
    setFunc(buf);
}

void Cell::setFunc(string const &funcText)
{
    // Clear any existing function, set the
    // new function, and update the cached
    // value:
    clear();
    formula.setText(funcText);

    // Subscribe to the sources of the function:
    list<string> srcs = formula.getSources();
    for(list<string>::iterator i = srcs.begin();
        i!=srcs.end();
        ++i)
        {
            Cell &src = container->at(*i);
            src.subscribe(this);
        }
    
    // Force an immediate update:
    update(NULL);
}

void Cell::clear()
{
    // Unsubscribe this cell from its
    // current sources:
    list<string> srcs = formula.getSources();
    for(list<string>::iterator i = srcs.begin();
        i!=srcs.end();
        ++i)
        {
            Cell * src = container->findCell(*i);
            if (src)
                src->unsubscribe(this);
        }
    // Set the function to a null value:
    formula.setText("");
    cachedValue = NULL_VALUE;
}

bool Cell::isNull() const
{
    return (formula.getText().empty());
}

void Cell::printOn(ostream &o, bool showFunc) const
{
    // Print out the current value
    // if there is one, else print a
    // default null string:
    if (!isNull())
    {
        if (showFunc)
        {
            string s = formula.getText();
            o << "=" << s ;
        }
        else
            o << cachedValue;
    }
    else
        o << "";
}


//----------------------------------------------
#pragma warning(disable:4355) // "this used in base member initializer list" -- incorrect warning
Spreadsheet::Spreadsheet()
 : theNullCell(this)
{}

Spreadsheet::Spreadsheet(Spreadsheet const &other)
{
    copyFrom(other);
}

Spreadsheet::~Spreadsheet()
{
    // Clear out the cells (resets inter-cell
    // references and unsubscribes cells from
    // each other; must do this before destroying
    // them)
    clear();
}


Spreadsheet &Spreadsheet::operator=(Spreadsheet const &other)
{
    copyFrom(other);
    return *this;
}

void Spreadsheet::copyFrom(Spreadsheet const &other)
{
    matrix = other.matrix;
    for(Matrix::iterator row = matrix.begin();
        row != matrix.end();
        ++row)
        {
            for(CellRow::iterator i = row->begin();
                i != row->end();
                ++i)
                {
                    // Make an independent copy:
                    if (*i)
                    {
                        *i = new Cell(**i);
                        (**i).setContainer(this);
                    }
                }
        }

}

void Spreadsheet::setSize(int rows,int cols)
{
    clear();
    ensureSize(rows+1, cols+1);
}

void Spreadsheet::clear()
{
    for(Matrix::iterator row = matrix.begin();
        row != matrix.end();
        ++row)
        {
            for(CellRow::iterator i = row->begin();
                i != row->end();
                ++i)
                {
                    delete (*i);
                    *i = NULL;
                }
        }
}

void Spreadsheet::printOn(ostream &o, bool showFuncs) const
{
    const int CELLWIDTH = 8;
    cout << "    ";
    for(char c='A';c<'I';++c)
        cout << setw(CELLWIDTH) << c;
    cout << endl;
    int rowNum = 0;
    for(Matrix::const_iterator row = matrix.begin();
        row != matrix.end();
        ++row)
        {
            o << setw(3) << ++rowNum << ":";
            for(CellRow::const_iterator i = row->begin();
                i != row->end();
                ++i)
                {
                    o << setw(CELLWIDTH);
                    if (*i)
                        (*i)->printOn(o, showFuncs);
                    else
                        o << " "; // Empty cell
                }
            o << endl;
        }
}

void Spreadsheet::ensureSize(int row, int col)
{
    int numRows = row;
    int numCols = col;

    // First, append enough new elements to each row
    // to make it (numCols) long:
    for(Matrix::iterator ri = matrix.begin();
        ri != matrix.end();
        ++ri)
        {
            while (ri->size() < numCols)
                ri->push_back(NULL);
        }

    // Next, append enough rows of size (numCols)
    // to make sure we're up to (numRows) rows:
    CellRow init(numCols, NULL); // Create a row #cols long
    while(matrix.size() < numRows)
        matrix.push_back(init); // Add a row

}

// Translates an "A1" style address into a row/column;
// returns true if cell currently exists:
bool Spreadsheet::translateAddress(std::string const &address,
                                   int &rowOut, int &colOut) const
{
    if (address.size() < 2) 
        throw BadCellRef(address);

    // Extract column number:
    char charCol = tolower(address[0]);
    if (!isalpha(charCol))
        throw BadCellRef(address);
    colOut = 1+charCol - 'a'; // US-ASCII only!

    // Extract row number:
    rowOut = atol(address.c_str()+1);
    if (rowOut==0)
        throw BadCellRef(address);

    assert(rowOut >= 1);
    assert(colOut >= 1);

    return ((matrix.size() >= rowOut) &&
            (matrix[rowOut-1].size() >= colOut));
}

// Find a cell, returning NULL if address does not exist
Cell const *Spreadsheet::findCell(std::string const &address) const
{
    int row,col;
    if (translateAddress(address,row,col))
        return &at_rowcol(row,col);
    else
        return NULL;
}

Cell *Spreadsheet::findCell(std::string const &address)
{
    int row,col;
    if (translateAddress(address,row,col))
        return &at_rowcol(row,col);
    else
        return NULL;
}

// Get access to a cell based on an "A1" style reference
Cell &Spreadsheet::at(std::string const &address)
{
    int row,col;
    translateAddress(address,row,col);
    ensureSize(row,col);
    return at_rowcol(row,col);
}   

Cell & Spreadsheet::at_rowcol(int row, int col) 
{
    // If the cell at the given address is empty,
    // create a null cell and return it.
    Cell *&cell = matrix[row-1][col-1];
    if (!cell) cell = new Cell(this);
    return *cell;
}

Cell const & Spreadsheet::at_rowcol(int row, int col) const 
{
    // If the cell at the given address is empty, return
    // a reference to a placeholder null cell.
    Cell *cell = matrix[row-1][col-1];
    if (!cell) return theNullCell;
    return *cell;
}

#if 0
void Spreadsheet::setCell(std::string const &address, auto_ptr<Cell> &cellPtr)
{
    int row,col;
    translateAddress(address,row,col);
    ensureSize(row,col);

    // Following operations cannot throw exceptions:
    Cell *currentCellPtr = matrix[row-1][col-1];
    delete currentCellPtr;
    matrix[row-1][col-1] = cellPtr.release();
    cellPtr->setContainer(this);
}
#endif
